import numpy as np
from numba import njit, prange


@njit(parallel=True, fastmath=True)
def _lnorm_multidimensional(x, y, weight, p=2):
    """

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.
    p: int
        Lp norm distance degree.

    Returns
    -------
        The Lp norm distance.
    """
    l1 = x.shape[0]
    l3 = x.shape[1]

    distance = np.zeros_like(x, dtype=float)
    for i in prange(l1):
        dist = 0.0
        for di in range(l3):
            diff = x[i, di] - y[i, di]
            dist += weight[i, di] * (diff ** p)
        distance[i] = dist ** (1 / p)

    return distance


def _lnorm_unidimensional(x, y, weight, p=2):
    """

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.
    p: int
        Lp norm distance degree.

    Returns
    -------
        The Lp norm distance.
    """
    distance = weight * np.power(np.power(np.abs(x - y), p), (1 / p))

    return distance



import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import novainstrumentation as ni

from matplotlib.collections import LineCollection


# Visualisation
def plot_alignment(ref_signal, estimated_signal, path, **kwargs):
    """
    This functions plots the resulted alignment of two sequences given the path
    calculated by the Dynamic Time Warping algorithm.

    :param ref_signal: (array-like)
                     The reference sequence.
    :param estimated_signal: (array-like)
                     The estimated sequence.
    :param path: (array-like)
                     A 2D array congaing the path resulted from the algorithm
    :param \**kwargs:
        See below:

        * *offset* (``double``) --
            The offset used to move the reference signal to an upper position for
            visualization purposes.
            (default: ``2``)

        * *linewidths* (``list``) --
            A list containing the linewidth for the reference, estimated and connection
            plots, respectively.
            (default: ``[3, 3, 0.5]``)

        * *step* (``int``) --
            The step for
          (default: ``2``)

        * *colors* (``list``) --
          A list containing the colors for the reference, estimated and connection
          plots, respectively.
          (default: ``[sns.color_palette()[0], sns.color_palette()[1], 'k']``)
    """

    step = kwargs.get("step", 2)
    hoffset = kwargs.get("hoffset", 0)
    voffset = kwargs.get("offset", 2)
    linewidths = kwargs.get("linewidths", [3, 3, 0.5])
    colors = kwargs.get("colors", [sns.color_palette()[0], sns.color_palette()[1], "k"])

    copy_ref = np.copy(ref_signal)  # This prevents unexpected changes in the reference signal after the duplicate
    copy_ref += voffset * np.max(ref_signal)  # Set an offset for visualization

    xref = np.arange(len(copy_ref)) + hoffset
    # Actual plot occurs here
    plt.plot(xref, copy_ref, color=sns.color_palette()[0], lw=linewidths[0], label="reference")
    plt.plot(estimated_signal, color=sns.color_palette()[1], lw=linewidths[1], label="estimate")
    plt.legend(fontsize=17)

    [
        plt.plot(
            [[path[0][i] + hoffset], [path[1][i]]],
            [copy_ref[path[0][i]], estimated_signal[path[1][i]]],
            color=colors[2],
            lw=linewidths[2],
        )
        for i in range(len(path[0]))[::step]
    ]

    ref_pks = ni.peakdelta(copy_ref, 0.5)
    est_pks = ni.peakdelta(estimated_signal, 0.5)

    if len(ref_pks[0]) > 0 and len(ref_pks[0]) == len(est_pks[0]):
        for i in range(len(ref_pks[0])):
            plt.plot([ref_pks[0][i, 0], est_pks[0][i, 0]], [ref_pks[0][i, 1], est_pks[0][i, 1]], "red")

    if len(ref_pks[1]) > 0 and len(ref_pks[1]) == len(est_pks[1]):
        for i in range(len(ref_pks[1])):
            plt.plot([ref_pks[1][i, 0], est_pks[1][i, 0]], [ref_pks[1][i, 1], est_pks[1][i, 1]], "red")


def plot_costmatrix(matrix, path):
    """
    This functions overlays the optimal warping path and the cost matrices
    :param matrix: (ndarray-like)
                The cost matrix (local cost or accumulated)
    :param path:   (ndarray-like)
                The optimal warping path
    :return: (void)
                Plots the optimal warping path with an overlay of the cost matrix.
    """
    plt.imshow(matrix.T, cmap="viridis", origin="lower", interpolation="None")
    plt.colorbar()
    plt.plot(path[0], path[1], "w.-")
    plt.xlim((-0.5, matrix.shape[0] - 0.5))
    plt.ylim((-0.5, matrix.shape[1] - 0.5))


def plot_search_distance_result(res, sequence, ts=None, cmap_name="viridis"):

    if ts is None:
        ts = np.arange(len(sequence))
    # set distance scale
    cmap = plt.cm.get_cmap(cmap_name)
    colors = cmap(np.arange(cmap.N))

    if len(np.shape(sequence)) > 1:
        sequence_shape = np.shape(sequence)[1]
    else:
        sequence_shape = 1

    all_axs = []
    for k in res.keys():
        max_dist = np.max(res[k]["path_dist"])
        min_dist = np.min(res[k]["path_dist"])
        if max_dist == min_dist:
            max_dist += 1
            min_dist -= 1
        delta_dist = max_dist - min_dist

        fig, axs = plt.subplots(sequence_shape + 1, 1, figsize=(15, 5))
        axs[0].set_title(k)
        for i in range(sequence_shape):
            plot_seq = sequence if sequence_shape == 1 else sequence[:, i]
            axs[i].plot(ts, plot_seq, "lightgray")
            for s, e, d in zip(res[k]["start"], res[k]["end"], res[k]["path_dist"]):
                d_idx = int((d - min_dist) * cmap.N / delta_dist) - 1
                axs[i].plot(ts[np.arange(s, e)], plot_seq[s:e], c=colors[d_idx])
            if i < sequence_shape - 2:
                axs[i].sharex(axs[i + 1])
                axs[i].set_xticks([])
                # [axs[i-1].sharex(axs[i]) for i in range(1, sequence_shape)]
        axs[sequence_shape].set_xlabel("Distance")
        axs[sequence_shape].imshow([colors], extent=[min_dist, max_dist, 0, 0.02 * delta_dist])
        axs[sequence_shape].set_yticks([])

        all_axs += [axs]

    return all_axs


def plot_weight_query(x, query, weight, cmap="viridis", axs=None, fig=None):

    points = np.array([x, query]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    if axs is None:
        fig = plt.figure()
        axs = plt.subplot()

    lc = LineCollection(segments, cmap=cmap, norm=plt.Normalize(0, weight.max()))
    lc.set_array(weight)
    lc.set_linewidth(2)
    line = axs.add_collection(lc)

    cbar = fig.colorbar(line, ax=axs)
    cbar.set_label("weight")

    axs.set_xlim(x.min(), x.max())
    axs.set_ylim(query.min() - 1, query.max() + 1)
    plt.show()



import numpy as np
from scipy.signal import find_peaks


def elastic_search(dict_distances, query, sequence, tq=None, ts=None, weight=None):
    """
    Query search for elastic measures

    Parameters
    ----------
    dict_distances: dict
        Configuration file with distances
    query: nd-array
        Query time series.
    sequence: nd-array
        Sequence time series.
    tq: nd-array
        Time stamp time series query.
    ts: nd-array
        Time stamp time series sequence.
    weight: nd-array (Default: None)
        query weight values

    Returns
    -------
    distance: nd-array
        distance value between query and sequence
    ac: nd-array
        accumulated cost matrix
    """

    exec("from tssearch import *")

    # distance function
    func_total = dict_distances["function"]

    # Check for parameters
    parameters_total = {}
    if dict_distances["parameters"] != "":
        parameters_total = dict_distances["parameters"]
    parameters_total["report"] = "search"

    if "dtw_type" in parameters_total:
        if parameters_total["dtw_type"] == "dtw":
            parameters_total["dtw_type"] = "sub-dtw"

    if "time" in parameters_total:
        parameters_total_copy = parameters_total.copy()
        del parameters_total_copy["time"]
        distances, ac = locals()[func_total](query, sequence, tq, ts, **parameters_total_copy)
    else:
        distances, ac = locals()[func_total](query, sequence, **parameters_total)

    return distances, ac


def lockstep_search(dict_distances, query, sequence, weight):
    """
    Query search for lockstep measures

    Parameters
    ----------
    dict_distances: dict
        Configuration file with distances
    query: nd-array
        Query time series.
    sequence: nd-array
        Sequence time series.
    weight: nd-array (Default: None)
        query weight values

    Returns
    -------
    res: nd-array
        distance value between query and sequence
    """

    exec("from tssearch import *")

    # distance function
    func_total = dict_distances["function"]

    # Check for parameters
    parameters_total = {}
    if dict_distances["parameters"] != "":
        parameters_total = dict_distances["parameters"]

    lw = len(query)
    res = np.zeros(len(sequence) - lw, "d")
    for i in range(len(sequence) - lw):
        seq_window = sequence[i : i + lw]

        eval_result = locals()[func_total](seq_window, query, weight, **parameters_total)

        res[i] = eval_result / lw  # default normalization

    return res


def start_sequences_index(distance, output=("number", 1), overlap=1.0):
    """
    Method to retrieve the k-best occurrences from a given vector distance

    Parameters
    ----------
    distance: nd-array
        distance values
    output: tuple
        number of occurrences
    overlap: float
        minimum distance between occurrences

    Returns
    -------
    id_s: nd-array
        indexes of k-best occurrences
    """

    # pks - min
    pks, _ = find_peaks(-distance, distance=overlap)  # TODO if necessary add first and last sequence
    pks_val = distance[pks]

    if output[0] == "number":
        num_events = output[1]
        pks_val_sort = np.argsort(pks_val)
        id_s = pks[pks_val_sort[:num_events]]
    elif output[0] == "percentile":
        perct = output[1]
        perct_val = np.percentile(distance, 100 - perct)
        pks_perct = np.where(pks_val < perct_val)[0]
        id_s = pks[pks_perct]
    elif output[0] == "threshold":
        thres = output[1]
        pks_thres = np.where(pks_val < thres)[0]
        id_s = pks[pks_thres]
    else:
        id_s = pks[np.argmin(pks_val)]

    return id_s



import numpy as np


def standardization(signal, fit=False, param=None):
    """Normalizes a given signal by subtracting the mean and dividing by the standard deviation.

    Parameters
    ----------
    signal : nd-array
        input signal

    Returns
    -------
    nd-array
        standardized signal

    """
    if param is not None:
        s_mean = param[0]
        s_std = param[1]
    else:
        s_mean = np.mean(signal, axis=0)
        s_std = np.std(signal, axis=0)

    if fit:
        d_mean = np.mean(np.diff(signal, axis=0), axis=0)
        d_std = np.std(np.diff(signal, axis=0), axis=0)
        return (signal - s_mean) / s_std, np.array([s_mean, s_std, d_mean, d_std])
    else:
        return (signal - s_mean) / s_std


def interpolation(x, y):
    """Computes the interpolation given two time series of different length.


    Parameters
    ----------
    x : nd-array
        Time series x
    y : nd-array
        Time series y

    Returns
    -------
    interp_signal (nd-array)
        Interpolated signal
    nd-array
        Time series

    """

    lx = len(x)
    ly = len(y)
    if lx > ly:
        t_old = np.linspace(0, lx, ly)
        t_new = np.linspace(0, lx, lx)
        if len(np.shape(x)) == 1:
            y_new = np.interp(t_new, t_old, y)
        else:
            y_new = np.array([np.interp(t_new, t_old, y[:, ax]) for ax in range(np.shape(x)[1])]).T
        return x, y_new
    else:
        t_old = np.linspace(0, ly, lx)
        t_new = np.linspace(0, ly, ly)

        if len(np.shape(x)) == 1:
            x_new = np.interp(t_new, t_old, x)
        else:
            x_new = np.array([np.interp(t_new, t_old, x[:, ax]) for ax in range(np.shape(x)[1])]).T
        return x_new, y



import numpy as np
import pandas as pd
from tssearch.search.segmentation import time_series_segmentation


def time_series_distance(dict_distances, x, y, tx=None, ty=None):
    """

    Parameters
    ----------
    dict_distances: dict
        Dictionary of distances parameters.
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    tx: nd-array
        Time stamp time series x.
    ty: nd-array
        Time stamp time series y.

    Returns
    -------
    distances: pandas DataFrame
        Distances values.

    """

    exec("from tssearch import *")

    distance_results = []
    distance_names = []

    multivariate = True if len(np.shape(x)) > 1 else False

    for d_type in dict_distances:
        for dist in dict_distances[d_type]:

            # Only returns used functions
            if "use" not in dict_distances[d_type][dist] or dict_distances[d_type][dist]["use"] == "yes":
                # remove unidimensional distances
                if multivariate and dict_distances[d_type][dist]["multivariate"] == "no":
                    continue

                func_total = dict_distances[d_type][dist]["function"]

                # Check for parameters
                parameters_total = {}
                if dict_distances[d_type][dist]["parameters"] != "":
                    parameters_total = dict_distances[d_type][dist]["parameters"]

                if "time" in parameters_total:
                    parameters_total_copy = parameters_total.copy()
                    del parameters_total_copy["time"]
                    eval_result = locals()[func_total](x, y, tx, ty, **parameters_total_copy)
                else:
                    eval_result = locals()[func_total](x, y, **parameters_total)

                distance_results += [eval_result]
                distance_names += [dist]

    distances = pd.DataFrame(data=np.array(distance_results), index=np.array(distance_names), columns=["Distance"])

    return distances


def time_series_distance_windows(dict_distances, x, y, tx=None, ty=None, segmentation=None):
    """

    Parameters
    ----------
    dict_distances: dict
        Dictionary of distances parameters.
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y (windows).
    tx: nd-array
        Time stamp time series x.
    ty: nd-array
        Time stamp time series y (windows).
    segmentation: dict
        Dictionary of distances parameters.

    Returns
    -------
    dist_windows: pandas DataFrame
        Distances values per window.

    """

    if segmentation is not None:
        results = time_series_segmentation(segmentation, x, y, tx, ty)
        func_name = list(segmentation[list(dict_distances.keys())[0]].keys())[0]

        ts_w = None if ty is None else []
        windows = []
        for i in range(len(results[func_name]) - 1):
            if ty is not None:
                ts_w += [ty[results[func_name][i] : results[func_name][i + 1]]]
            windows += [y[results[func_name][i] : results[func_name][i + 1]]]
    else:
        windows = y
        ts_w = ty

    multivariate = True if len(np.shape(x)) > 1 else False

    exec("from tssearch import *")

    dist_windows = pd.DataFrame()
    for d_type in dict_distances:
        for dist in dict_distances[d_type]:

            # Only returns used functions
            if "use" not in dict_distances[d_type][dist] or dict_distances[d_type][dist]["use"] == "yes":

                if multivariate and dict_distances[d_type][dist]["multivariate"] == "no":
                    continue

                func_total = dict_distances[d_type][dist]["function"]

                # Check for parameters
                parameters_total = {}
                if dict_distances[d_type][dist]["parameters"] != "":
                    parameters_total = dict_distances[d_type][dist]["parameters"]

                distance_results = []
                if "time" in parameters_total:
                    parameters_total_copy = parameters_total.copy()
                    del parameters_total_copy["time"]
                    for ty_window, window in zip(ts_w, windows):
                        eval_result = locals()[func_total](x, window, tx, ty_window, **parameters_total_copy)
                        distance_results += [eval_result]
                else:
                    for window in windows:
                        eval_result = locals()[func_total](x, window, **parameters_total)
                        distance_results += [eval_result]

                dist_windows[dist] = distance_results

    return dist_windows



import json
import tssearch


def load_json(json_path):
    """Loads the json file given by filename.
    Parameters
    ----------
    json_path : string
        Json path
    Returns
    -------
    Dict
        Dictionary
    """

    return json.load(open(json_path))


def get_distances_by_type(domain=None, json_path=None):
    """Creates a dictionary with the features settings by domain.
    Parameters
    ----------
    domain : string
        Available domains: "statistical"; "spectral"; "temporal"
        If domain equals None, then the features settings from all domains are returned.
    json_path : string
        Directory of json file. Default: package features.json directory
    Returns
    -------
    Dict
        Dictionary with the features settings
    """

    if json_path is None:
        json_path = tssearch.__path__[0] + "/distances/distances.json"

        if domain not in ["elastic", "lockstep", "time", None]:
            raise SystemExit("No valid domain. Choose: lockstep, elastic, time or None (for all distances settings).")

    dict_features = load_json(json_path)
    if domain is None:
        return dict_features
    else:
        return {domain: dict_features[domain]}


def get_distance_dict(dist_list):

    json_path = tssearch.__path__[0] + "/distances/distances.json"

    dict_features = load_json(json_path)

    select_distances = {}
    for d in dist_list:
        if d in dict_features["elastic"]:
            d_type = "elastic"
        elif d in dict_features["lockstep"]:
            d_type = "lockstep"
        elif d in dict_features["time"]:
            d_type = "time"
        else:
            continue

        if d_type not in select_distances:
            select_distances[d_type] = {}
        select_distances[d_type][d] = dict_features[d_type][d]

    return select_distances



import numpy as np
from scipy import stats
from scipy.spatial import distance
from tssearch.utils.preprocessing import interpolation
from tssearch.distances.lockstep_utils import _lnorm_multidimensional, _lnorm_unidimensional


def euclidean_distance(x, y, weight=None):
    """Computes the Euclidean distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Euclidean distance value.

    """
    p = 2

    if len(x) != len(y):
        x, y = interpolation(x, y)

    if weight is None:
        ed = np.linalg.norm(x - y, p)
    else:
        if len(np.shape(x)) > 1:
            distance = _lnorm_multidimensional(x, y, weight, p=p)
        else:
            distance = _lnorm_unidimensional(x, y, weight, p=p)
        ed = np.sum(distance)
    return ed


def minkowski_distance(x, y, weight=None, p=3):
    """Computes the Minkowski distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.
    p: int
        Lp norm distance degree.

    Returns
    -------
    float
        Minkowski distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    if weight is None and (p < 3 or p == np.inf):
        distance = np.linalg.norm(x - y, p)
    else:
        if weight is None:
            weight = np.ones_like(x)
        if len(np.shape(x)) > 1:
            distance = _lnorm_multidimensional(x, y, weight, p=p)
        else:
            distance = _lnorm_unidimensional(x, y, weight, p=p)
        distance = np.sum(distance)

    return distance


def manhattan_distance(x, y, weight=None):
    """Computes the Manhattan distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Manhattan distance value.

    """
    p = 1

    if len(x) != len(y):
        x, y = interpolation(x, y)

    if weight is None:
        distance = np.linalg.norm(x - y, p)
    else:
        if len(np.shape(x)) > 1:
            distance = _lnorm_multidimensional(x, y, weight, p=p)
        else:
            distance = _lnorm_unidimensional(x, y, weight, p=p)
        distance = np.sum(distance)

    return distance


def chebyshev_distance(x, y, weight=None):
    """Computes the Chebyshev distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Chebyshev distance value.

    """
    p = np.inf

    if len(x) != len(y):
        x, y = interpolation(x, y)

    if weight is None:
        d = np.linalg.norm(x - y, p)
    else:
        if len(np.shape(x)) > 1:
            distance = _lnorm_multidimensional(x, y, weight, p=p)
        else:
            distance = _lnorm_unidimensional(x, y, weight, p=p)
        d = np.sum(distance)
    return d


def correlation_distance(x, y, weight=None):
    """Computes the correlation distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Correlation distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    correlation_d = distance.correlation(x, y, weight)

    return correlation_d


def pearson_correlation(x, y, beta=None):
    """Computes the Pearson correlation between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    beta: float
        Beta coefficient.

    Returns
    -------
    float
        Pearson correlation value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    r, p = stats.pearsonr(x, y)

    if beta is None:
        d = 2 * (1 - r)
    else:
        d = ((1 - r) / (1 + r)) ** beta
    return d


def short_time_series_distance(x, y, tx=None, ty=None):
    """Computes the short time series distance (STS) between two time series.

    Reference: Möller-Levet, C. S., Klawonn, F., Cho, K., and Wolkenhauer, O. (2003).

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    tx : nd-array
        Sampling index of time series x.
    ty : nd-array
        Sampling index of time series y.

    Returns
    -------
    float
        Short time series distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    if tx is None:
        tx = np.arange(len(x))
    if ty is None:
        ty = np.arange(len(y))

    sts = np.sqrt(np.sum((np.diff(y) / np.diff(tx) - np.diff(x) / np.diff(ty)) ** 2))

    return sts


def braycurtis_distance(x, y, weight=None):
    """Computes the Braycurtis distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Braycurtis distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    braycurtis_d = distance.braycurtis(x, y, weight)

    return braycurtis_d


def canberra_distance(x, y, weight=None):
    """Computes the Canberra distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Canberra distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    canberra_d = distance.canberra(x, y, weight)

    return canberra_d


def cosine_distance(x, y, weight=None):
    """Computes the cosine distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Cosine distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    cosine_d = distance.cosine(x, y, weight)

    return cosine_d


def mahalanobis_distance(x, y, weight=None):
    """Computes the Mahalanobis distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Mahalanobis distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    mahalanobis_d = distance.mahalanobis(x, y, weight)

    return mahalanobis_d


def sqeuclidean_distance(x, y, weight=None):
    """Computes the squared Euclidean distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Squared Euclidean distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    sqeuclidean_d = distance.sqeuclidean(x, y, weight)

    return sqeuclidean_d


def hamming_distance(x, y, weight=None):
    """Computes the Hamming distance between two time series.

    If the time series do not have the same length, an interpolation is performed.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.
    weight: nd-array (Default: None)
        query weight values.

    Returns
    -------
    float
        Hamming distance value.

    """
    if len(x) != len(y):
        x, y = interpolation(x, y)

    hamming_d = distance.hamming(x, y, weight)

    return hamming_d



import numpy as np

from tssearch.distances.elastic_utils import (
    cost_matrix,
    accumulated_cost_matrix,
    acc_initialization,
    lcss_accumulated_matrix,
    lcss_path,
    lcss_score,
    traceback_adj,
    backtracking,
)


def dtw(x, y, weight=None, **kwargs):
    """Computes Dynamic Time Warping (DTW) of two time series.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    dist: function
        The distance used as a local cost measure. None defaults to the squared euclidean distance.
    \**kwargs:
    See below:

    * *do_sign_norm* (``bool``) --
      If ``True`` the signals will be normalized before computing the DTW,
      (default: ``False``)

    * *do_dist_norm* (``bool``) --
      If ``True`` the DTW distance will be normalized by dividing the summation of the path dimension.
      (default: ``True``)

    * *window* (``String``) --
      Selects the global window constrains. Available options are ``None`` and ``sakoe-chiba``.
      (default: ``None``)

    * *factor* (``Float``) --
      Selects the global constrain factor.
      (default: ``min(xl, yl) * .50``)


    Returns
    -------
    d: float
        The DTW distance.
    ac: nd-array
        The accumulated cost matrix.
    path: nd-array
        The optimal warping path between the two sequences.
    """

    xl, yl = len(x), len(y)

    alpha = kwargs.get("alpha", 1)
    do_dist_norm = kwargs.get("dist_norm", True)
    window = kwargs.get("window", None)
    factor = kwargs.get("factor", np.min((xl, yl)) * 0.50)
    dtw_type = kwargs.get("dtw_type", "dtw")
    tolerance = kwargs.get("tolerance", 0)
    report = kwargs.get("report", "distance")

    # cost matrix
    c = cost_matrix(x, y, alpha, weight=weight)
    # Acc cost matrix
    ac = accumulated_cost_matrix(c, window=window, factor=factor, dtw_type=dtw_type, tolerance=tolerance)

    # Distance
    if report == "cost_matrix":
        return ac
    elif report == "search":
        d = ac[-1, :]
        return d, ac
    elif report == "path":
        path = traceback_adj(ac)
        return path
    else:  # report = "distance" default
        d = ac[-1, -1] / xl if do_dist_norm else ac[-1, -1]
        return d


def lcss(x, y, eps=1, **kwargs):
    """Computes the Longest Common Subsequence (LCSS) distance between two numeric time series.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    eps : float
        Amplitude matching threshold.
    \**kwargs:
    See below:

    * *window* (``String``) --
      Selects the global window constrains. Available options are ``None`` and ``sakoe-chiba``.
      (default: ``None``)

    Returns
    -------
    d: float
        The LCSS distance.
    ac: nd-array
        The similarity matrix.
    path: nd-array
        The optimal path between the two sequences.
    """

    window = kwargs.get("window", None)
    report = kwargs.get("report", "distance")

    dim = len(np.shape(x))  # tem de dar erro se forem inseridas duas TS com dims diferentes
    if dim == 1:
        x = x.reshape(-1, 1)
        y = y.reshape(-1, 1)

    ac = lcss_accumulated_matrix(x, y, eps=eps)
    path = lcss_path(x, y, ac, eps=eps)
    sim_score = lcss_score(ac)

    if report == "cost_matrix":
        return ac
    elif report == "search":
        return sim_score, ac
    elif report == "path":
        return path
    else:
        return sim_score


def dlp(x, y, p=2):
    """Computes Lp norm distance between two time series.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    p: int
        Lp norm distance degree for local cost computation.

    Returns
    -------
        The Lp distance.
    """

    cost = np.sum(np.power(np.abs(x - y), p))
    return np.power(cost, 1 / p)


def twed(x, y, tx, ty, nu=0.001, lmbda=1.0, p=2, report="distance"):
    """Computes Time Warp Edit Distance (TWED) of two time series.

    Reference :
       Marteau, P.; F. (2009). "Time Warp Edit Distance with Stiffness Adjustment for Time Series Matching".
       IEEE Transactions on Pattern Analysis and Machine Intelligence. 31 (2): 306–318. arXiv:cs/0703033
       http://people.irisa.fr/Pierre-Francois.Marteau/

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    tx: nd-array
        Time stamp time series x.
    ty: nd-array
        Time stamp time series y.
    nu: int
        Stiffness parameter (nu >= 0)
            nu = 0, TWED distance measure on amplitude.
            nu > 0, TWED distance measure on amplitude x time.
    lmbda: int
        Penalty for deletion operation (lmbda >= 0).
    p: int
        Lp norm distance degree for local cost computation.
    report: str
        distance, cost matrix, path.

    Returns
    -------
    d: float
        The TWED distance.
    ac: nd-array
        The accumulated cost matrix.
    path: nd-array
        The optimal warping path between the two sequences.
    """

    # Check if input arguments
    if len(x) != len(tx):
        print("The length of x is not equal length of tx")
        return None, None

    if len(y) != len(ty):
        print("The length of y is not equal length of ty")
        return None, None

    if nu < 0:
        print("nu is negative")
        return None, None

    # Dynamical programming
    ac = acc_initialization(len(x), len(y), report)

    # Add padding
    query = np.array([0] + list(x))
    tq = np.array([0] + list(tx))
    sequence = np.array([0] + list(y))
    ts = np.array([0] + list(ty))

    n = len(query)
    m = len(sequence)

    # Compute minimal cost
    for i in range(1, n):
        for j in range(1, m):
            # Calculate and save cost of various operations
            C = np.ones((3, 1)) * np.inf
            # Deletion in A
            C[0] = ac[i - 1, j] + dlp(query[i - 1], query[i], p) + nu * (tq[i] - tq[i - 1]) + lmbda
            # Deletion in B
            C[1] = ac[i, j - 1] + dlp(sequence[j - 1], sequence[j], p) + nu * (ts[j] - ts[j - 1]) + lmbda
            # Keep data points in both time series
            C[2] = (
                ac[i - 1, j - 1]
                + dlp(query[i], sequence[j], degree)
                + dlp(query[i - 1], sequence[j - 1], p)
                + nu * (abs(tq[i] - ts[j]) + abs(tq[i - 1] - ts[j - 1]))
            )
            # Choose the operation with the minimal cost and update c Matrix
            ac[i, j] = np.min(C)

    if report == "cost_matrix":
        return ac
    elif report == "search":
        d = ac[n - 1, :]
        return d, ac
    elif report == "path":
        path = backtracking(ac)
        return path
    else:  # report = 'search'
        return ac[n - 1, m - 1]



from scipy.signal import find_peaks
from tssearch.search.search_utils import lockstep_search, elastic_search


def time_series_segmentation(dict_distances, query, sequence, tq=None, ts=None, weight=None):
    """
    Time series segmentation locates the time instants between consecutive query repetitions on a more extended and
    repetitive sequence.

    Parameters
    ----------
    dict_distances: dict
        Configuration file with distances
    query: nd-array
        Query time series.
    sequence: nd-array
        Sequence time series.
    tq: nd-array
        Time stamp time series query.
    ts: nd-array
        Time stamp time series sequence.
    weight: nd-array (Default: None)
        query weight values
    Returns
    -------
    segment_results: dict
        Segmented time instants for each given distances
    """

    l_query = len(query)
    segment_results = {}

    for d_type in dict_distances:
        for dist in dict_distances[d_type]:

            if "use" not in dict_distances[d_type][dist] or dict_distances[d_type][dist]["use"] == "yes":
                segment_results[dist] = {}
                if d_type == "lockstep":
                    distance = lockstep_search(dict_distances[d_type][dist], query, sequence, weight)
                elif d_type == "elastic":
                    distance, ac = elastic_search(dict_distances[d_type][dist], query, sequence, tq, ts, weight)
                else:
                    print("WARNING")
                    continue

                pks, _ = find_peaks(-distance, distance=l_query / 2)
                segment_results[dist] = pks

    return segment_results



from tssearch.distances.lockstep_distances import *
from tssearch.distances.elastic_distances import *
from tssearch.distances.lockstep_utils import *
from tssearch.distances.elastic_utils import *
from tssearch.distances.time_distances import *
from tssearch.distances.compute_distance import *



import tssearch
import pickle
import numpy as np


def load_ecg_example():

    filename = tssearch.__path__[0] + "/examples/ecg.pickle"
    with open(filename, "rb") as handle:
        data = pickle.load(handle)

    return data



import numpy as np
from numba import njit, prange
from tssearch.utils.preprocessing import standardization


@njit(parallel=True, fastmath=True)
def _cost_matrix(x, y):
    """

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.

    Returns
    -------
    c: nd-array
        The cost matrix.
    """
    l1 = x.shape[0]
    l2 = y.shape[0]
    c = np.zeros((l1, l2), dtype=np.float32)

    for i in prange(l1):
        for j in prange(l2):
            c[i, j] = (x[i] - y[j]) ** 2

    return c


@njit(parallel=True, fastmath=True)
def _multidimensional_cost_matrix(x, y, weight):
    """Helper function for fast computation of cost matrix in cost_matrix_diff_vec.
    Defined outside to prevent recompilation from numba

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.

    Returns
    -------
    c: nd-array
        The cost matrix.
    """
    l1 = x.shape[0]
    l2 = y.shape[0]
    l3 = x.shape[1]
    c = np.zeros((l1, l2), dtype=np.float32)

    for i in prange(l1):
        for j in prange(l2):
            dist = 0.0
            for di in range(l3):
                diff = x[i, di] - y[j, di]
                dist += weight[i, di] * (diff * diff)
            c[i, j] = dist ** 0.5

    return c


@njit(nogil=True, fastmath=True)
def _accumulated_cost_matrix(ac):
    """Fast computation of accumulated cost matrix using cost matrix.

    Parameters
    ----------
    ac: nd-array
        Given cost matrix c, ac = acc_initialization(...), ac[1:, 1:] = c.

    Returns
    -------
        The accumulated cost matrix.
    """
    for i in range(ac.shape[0] - 1):
        for j in range(ac.shape[1] - 1):
            ac[i + 1, j + 1] += min(ac[i, j + 1], ac[i + 1, j], ac[i, j])
    return ac


def acc_initialization(x, y, _type, tolerance=0):
    """Initializes the cost matrix according to the dtw type.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    _type: string
        Name of dtw type.
    tolerance: int
        Tolerance value.

    Returns
    -------
    ac: nd-array
        The accumulated cost matrix.
    """
    ac = np.zeros((x + 1, y + 1))
    if _type == "dtw":
        ac[0, 1:] = np.inf
        ac[1:, 0] = np.inf
    elif _type == "oe-dtw":
        ac[0, 1:] = np.inf
        ac[1:, 0] = np.inf
    elif _type == "obe-dtw" or _type == "sub-dtw" or _type == "search":
        ac[1:, 0] = np.inf
    elif _type == "psi-dtw":
        ac[0, tolerance + 1 :] = np.inf
        ac[tolerance + 1 :, 0] = np.inf
    else:
        ac[0, 1:] = np.inf
        ac[1:, 0] = np.inf

    return ac


def cost_matrix(x, y, alpha=1, weight=None):
    """Computes cost matrix using a specified distance (dist) between two time series.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    dist: function
        The distance used as a local cost measure. None defaults to the squared euclidean distance.

    \**kwargs:
    See below:

    * *do_sign_norm* (``bool``) --
      If ``True`` the signals will be normalized before computing the DTW,
      (default: ``False``)

    * *do_dist_norm* (``bool``) --
      If ``True`` the DTW distance will be normalized by dividing the summation of the path dimension.
      (default: ``True``)

    * *window* (``String``) --
      Selects the global window constrains. Available options are ``None`` and ``sakoe-chiba``.
      (default: ``None``)

    * *factor* (``Float``) --
      Selects the global constrain factor.
      (default: ``min(xl, yl) * .50``)

    Returns
    -------
    c: nd-array
        The cost matrix.
    """
    if weight is None:
        weight = np.ones_like(x)

    if len(np.shape(weight)) == 1:
        weight = weight.reshape(-1, 1)

    if alpha == 1:
        C_d = 1
        if len(np.shape(x)) == 1:
            C_n = (_cost_matrix(x, y) * weight) / np.max(weight)
        else:
            C_n = _multidimensional_cost_matrix(x, y, weight)
    else:
        # standardization parameters
        abs_norm = np.mean(x, axis=0), np.std(x, axis=0)
        diff_norm = np.mean(np.diff(x, axis=0), axis=0), np.std(np.diff(x, axis=0), axis=0)

        # Derivative calculation and standardization
        _x = standardization(np.diff(x, axis=0), param=diff_norm)
        _y = standardization(np.diff(y, axis=0), param=diff_norm)
        # same length of derivative
        x = standardization(x[:-1], param=abs_norm)
        y = standardization(y[:-1], param=abs_norm)

        weight = weight[:-1]

        if len(np.shape(x)) == 1:
            C_d = _cost_matrix(_x, _y) * weight
            C_n = _cost_matrix(x, y) * weight
        else:
            C_d = _multidimensional_cost_matrix(_x, _y, weight)
            C_n = _multidimensional_cost_matrix(x, y, weight)

    c = alpha * C_n + (1 - alpha) * C_d

    return c


def accumulated_cost_matrix(c, **kwargs):
    """

    Parameters
    ----------
    c: nd-array
        The cost matrix.

    \**kwargs:

    Returns
    -------
    ac: nd-array
        The accumulated cost matrix.
    """
    xl, yl = np.shape(c)

    window = kwargs.get("window", None)
    factor = kwargs.get("factor", np.min((xl, yl)) * 0.50)
    dtw_type = kwargs.get("dtw_type", "dtw")
    tolerance = kwargs.get("tolerance", 0)

    if window == "sakoe-chiba":
        c[np.abs(np.diff(np.indices(c.shape), axis=0))[0] > factor] = np.inf

    ac = acc_initialization(xl, yl, dtw_type, tolerance)
    ac[1:, 1:] = c.copy()
    ac = _accumulated_cost_matrix(ac)[1:, 1:]

    return ac


@njit(nogil=True, fastmath=True)
def traceback(ac):
    """Computes the traceback path of the matrix c.

    Parameters
    ----------
    ac: nd-array
        The accumulated cost matrix.

    Returns
    -------
        Coordinates p and q of the minimum path.

    """

    i, j = np.array(ac.shape) - 2
    p, q = [i], [j]
    while (i > 0) and (j > 0):
        tb = 0
        if ac[i, j + 1] < ac[i, j]:
            tb = 1
        if ac[i + 1, j] < ac[i, j + tb]:
            tb = 2
        if tb == 0:
            i -= 1
            j -= 1
        elif tb == 1:
            i -= 1
        else:
            j -= 1
        p.insert(0, i)
        q.insert(0, j)
    while j > 0:
        j -= 1
        p.insert(0, i)
        q.insert(0, j)
    while i > 0:
        i -= 1
        p.insert(0, i)
        q.insert(0, j)

    return np.array(p), np.array(q)


@njit(nogil=True, fastmath=True)
def traceback_adj(ac):
    """Computes the adjusted traceback path of the matrix c.

    Parameters
    ----------
    ac: nd-array
        The accumulated cost matrix.

    Returns
    -------
        Coordinates p and q of the minimum path adjusted.

    """
    i, j = np.array(ac.shape) - 2
    p, q = [i], [j]
    while (i > 0) and (j > 0):
        tb = 0
        if ac[i, j + 1] < ac[i, j]:
            tb = 1
        if ac[i + 1, j] < ac[i, j + tb]:
            tb = 2
        if tb == 0:
            i -= 1
            j -= 1
        elif tb == 1:
            i -= 1
        else:  # (tb == 2):
            j -= 1
        p.insert(0, i)
        q.insert(0, j)
    while i > 0:
        i -= 1
        p.insert(0, i)
        q.insert(0, j)
    return np.array(p), np.array(q)


def backtracking(ac):
    """Compute the most cost-efficient path.

    Parameters
    ----------
    ac: nd-array
        The accumulated cost matrix.

    Returns
    -------
         Coordinates of the most cost-efficient path.
    """
    x = np.shape(ac)
    i = x[0] - 1
    j = x[1] - 1

    # The indices of the paths are save in opposite direction
    # path = np.ones((i + j, 2 )) * np.inf;
    best_path = []

    steps = 0
    while i != 0 or j != 0:

        best_path.append((i - 1, j - 1))

        C = np.ones((3, 1)) * np.inf

        # Keep data points in both time series
        C[0] = ac[i - 1, j - 1]
        # Deletion in A
        C[1] = ac[i - 1, j]
        # Deletion in B
        C[2] = ac[i, j - 1]

        # Find the index for the lowest cost
        idx = np.argmin(C)

        if idx == 0:
            # Keep data points in both time series
            i = i - 1
            j = j - 1
        elif idx == 1:
            # Deletion in A
            i = i - 1
            j = j
        else:
            # Deletion in B
            i = i
            j = j - 1
        steps = steps + 1

    best_path.append((i - 1, j - 1))

    best_path.reverse()
    best_path = np.array(best_path[1:])

    return best_path[:, 0], best_path[:, 1]


# DTW SW
def dtw_sw(x, y, winlen, alpha=0.5, **kwargs):
    """Computes Dynamic Time Warping (DTW) of two time series using a sliding window.
    TODO: Check if this needs to be speed up.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    winlen: int
        The sliding window length.
    alpha: float
        A factor between 0 and 1 which weights the amplitude and derivative contributions.
        A higher value will favor amplitude and a lower value will favor the first derivative.

    \**kwargs:
        See below:

        * *do_sign_norm* (``bool``) --
          If ``True`` the signals will be normalized before computing the DTW,
          (default: ``False``)

        * *do_dist_norm* (``bool``) --
          If ``True`` the DTW distance will be normalized by dividing the summation of the path dimension.
          (default: ``True``)

        * *window* (``String``) --
          Selects the global window constrains. Available options are ``None`` and ``sakoe-chiba``.
          (default: ``None``)

        * *factor* (``Float``) --
          Selects the global constrain factor.
          (default: ``min(xl, yl) * .50``)


    Returns
    -------
    d: float
        The SW-DTW distance.
    c: nd-array
        The local cost matrix.
    ac: nd-array
        The accumulated cost matrix.
    path: nd-array
        The optimal warping path between the two sequences.

    """
    xl, yl = len(x), len(y)

    do_sign_norm = kwargs.get("normalize", False)
    do_dist_norm = kwargs.get("dist_norm", True)
    window = kwargs.get("window", None)
    factor = kwargs.get("factor", np.min((xl, yl)) * 0.50)

    if do_sign_norm:
        x, y = standardization(x), standardization(y)

    ac = np.zeros((xl + 1, yl + 1))
    ac[0, 1:] = np.inf
    ac[1:, 0] = np.inf
    tmp_ac = ac[1:, 1:]

    nx = get_mirror(x, winlen)
    ny = get_mirror(y, winlen)

    dnx = np.diff(nx)
    dny = np.diff(ny)

    nx = nx[:-1]
    ny = ny[:-1]

    # Workaround to deal with even window sizes
    if winlen % 2 == 0:
        winlen -= 1

    swindow = np.hamming(winlen)
    swindow = swindow / np.sum(swindow)

    for i in range(xl):
        for j in range(yl):
            pad_i, pad_j = i + winlen, j + winlen
            # No window selected
            if window is None:
                tmp_ac[i, j] = sliding_dist(
                    nx[pad_i - (winlen // 2) : pad_i + (winlen // 2) + 1],
                    ny[pad_j - (winlen // 2) : pad_j + (winlen // 2) + 1],
                    dnx[pad_i - (winlen // 2) : pad_i + (winlen // 2) + 1],
                    dny[pad_j - (winlen // 2) : pad_j + (winlen // 2) + 1],
                    alpha,
                    swindow,
                )

            # Sakoe-Chiba band
            elif window == "sakoe-chiba":
                if abs(i - j) < factor:
                    tmp_ac[i, j] = sliding_dist(
                        nx[pad_i - (winlen // 2) : pad_i + (winlen // 2) + 1],
                        ny[pad_j - (winlen // 2) : pad_j + (winlen // 2) + 1],
                        dnx[pad_i - (winlen // 2) : pad_i + (winlen // 2) + 1],
                        dny[pad_j - (winlen // 2) : pad_j + (winlen // 2) + 1],
                        alpha,
                        swindow,
                    )
                else:
                    tmp_ac[i, j] = np.inf

            # As last resource, the complete window is calculated
            else:
                tmp_ac[i, j] = sliding_dist(
                    nx[pad_i - (winlen / 2) : pad_i + (winlen / 2) + 1],
                    ny[pad_j - (winlen / 2) : pad_j + (winlen / 2) + 1],
                    dnx[pad_i - (winlen / 2) : pad_i + (winlen / 2) + 1],
                    dny[pad_j - (winlen / 2) : pad_j + (winlen / 2) + 1],
                    alpha,
                    swindow,
                )

    c = tmp_ac.copy()

    for i in range(xl):
        for j in range(yl):
            tmp_ac[i, j] += min([ac[i, j], ac[i, j + 1], ac[i + 1, j]])

    path = traceback(ac)

    if do_dist_norm:
        d = ac[-1, -1] / np.sum(np.shape(path))
    else:
        d = ac[-1, -1]

    return d, c, ac, path


def sliding_dist(xw, yw, dxw, dyw, alpha, win):
    """Computes the sliding distance.

    Parameters
    ----------
    xw: nd-array
        x coords window.
    yw: nd-array
        y coords window.
    dxw: nd-array
        x coords diff window.
    dyw: nd-array
        y coords diff window.
    alpha: float
        Rely more on absolute or difference values 1- abs, 0 - diff.
    win: nd-array
        Signal window used for sliding distance.

    Returns
    -------
        Sliding distance
    """
    return (1 - alpha) * np.sqrt(np.sum((((dxw - dyw) * win) ** 2.0))) + alpha * np.sqrt(
        np.sum((((xw - yw) * win) ** 2.0))
    )


def get_mirror(s, ws):
    """Performs a signal windowing based on a double inversion from the start and end segments.

    Parameters
    ----------
    s: nd-array
            the input-signal.
    ws: int
            window size.

    Returns
    -------
        Signal windowed
    """

    return np.r_[2 * s[0] - s[ws:0:-1], s, 2 * s[-1] - s[-2 : -ws - 2 : -1]]


@njit()
def _lcss_point_dist(x, y):
    """

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.

    Returns
    -------
        The LCSS distance.
    """
    dist = 0.0
    for di in range(x.shape[0]):
        diff = x[di] - y[di]
        dist += diff * diff

    return dist ** 0.5


def lcss_accumulated_matrix(x, y, eps):
    """Computes the LCSS cost matrix using the euclidean distance (dist) between two time series.

    Parameters
    ----------
    x: nd-array
        Time series x (query).
    y: nd-array
        Time series y.
    eps : float
        Amplitude matching threshold.

    Returns
    -------
    ac : nd-array
            The accumulated cost matrix.
    """

    xl, yl = len(x), len(y)

    ac = np.zeros((xl + 1, yl + 1))

    for i in range(1, xl + 1):
        for j in range(1, yl + 1):
            if _lcss_point_dist(x[i - 1, :], y[j - 1, :]) <= eps:
                ac[i, j] = 1 + ac[i - 1, j - 1]
            else:
                ac[i, j] = max(ac[i, j - 1], ac[i - 1, j])

    return ac


def lcss_path(x, y, c, eps):
    """Computes the LCSS path between two time series.

    Parameters
    ----------
    x: nd-array
        The reference signal.
    y: nd-array
        The estimated signal.
    c : nd-array
        The cost matrix.
    eps : float
        Matching threshold.

    Returns
    -------
        Coordinates of the minimum LCSS path.
    """
    i, j = len(x), len(y)
    path = []

    while i > 0 and j > 0:
        if _lcss_point_dist(x[i - 1, :], y[j - 1, :]) <= eps:
            path.append((i - 1, j - 1))
            i -= 1
            j -= 1
        elif c[i - 1, j] > c[i, j - 1]:
            i -= 1
        else:
            j -= 1

    path = np.array(path[::-1])
    return path[1:, 0], path[1:, 1]


def lcss_score(c):
    """Computes the LCSS similarity score between two time series.

    Parameters
    ----------
    c : nd-array
        The cost matrix.

    Returns
    -------
        The LCSS score.
    """

    xl = c.shape[0] - 1
    yl = c.shape[1] - 1

    return float(c[-1, -1]) / min([xl, yl])



import numpy as np
from tssearch.distances.elastic_utils import traceback_adj, lcss_path, lcss_score
from tssearch.search.search_utils import lockstep_search, elastic_search, start_sequences_index


def time_series_search(dict_distances, query, sequence, tq=None, ts=None, weight=None, output=("number", 1)):
    """
    Time series search method locates the k-best occurrences of a given query on a more extended sequence based on a
    distance measurement.

    Parameters
    ----------
    dict_distances: dict
        Configuration file with distances.
    query: nd-array
        Query time series.
    sequence: nd-array
        Sequence time series.
    tq: nd-array
        Time stamp time series query.
    ts: nd-array
        Time stamp time series sequence.
    weight: nd-array (Default: None)
        query weight values.
    output: tuple
        number of occurrences.

    Returns
    -------
    distance_results: dict
        time instants, optimal alignment path and distance for each occurrence per distance.
    """

    l_query = len(query)
    distance_results = {}

    for d_type in dict_distances:
        for dist in dict_distances[d_type]:

            if "use" not in dict_distances[d_type][dist] or dict_distances[d_type][dist]["use"] == "yes":
                distance_results[dist] = {}
                if d_type == "lockstep":
                    distance = lockstep_search(dict_distances[d_type][dist], query, sequence, weight)

                    start_index = start_sequences_index(distance, output=output, overlap=l_query)
                    end_index, path = [], []
                    for start in start_index:
                        end_index += [start + l_query]
                        path += [(np.arange(l_query), np.arange(start, end_index[-1]))]
                    distance_results[dist]["path_dist"] = distance[start_index]
                elif d_type == "elastic":
                    distance, ac = elastic_search(dict_distances[d_type][dist], query, sequence, tq, ts, weight)

                    if dist == "Longest Common Subsequence":
                        eps = dict_distances[d_type][dist]["parameters"]["eps"]
                        if len(np.shape(query)) == 1:
                            query_copy = query.reshape(-1, 1)
                            sequence_copy = sequence.reshape(-1, 1)
                            path = [lcss_path(query_copy, sequence_copy, ac, eps)]
                        else:
                            path = [lcss_path(query, sequence, ac, eps)]
                        distance_results[dist]["path_dist"] = [lcss_score(ac)]
                        end_index = [path_i[1][-1] for path_i in path]
                    else:
                        end_index = start_sequences_index(distance, output=output, overlap=l_query / 2)
                        # check if traceback_adj is equal to other elastic measures
                        path = [traceback_adj(ac[:, : int(pk) + 1]) for pk in end_index]
                        distance_results[dist]["path_dist"] = distance[end_index]
                    start_index = [path_i[1][0] for path_i in path]

                else:
                    print("WARNING")
                    continue

                distance_results[dist]["distance"] = distance
                distance_results[dist]["start"] = start_index
                distance_results[dist]["end"] = end_index
                distance_results[dist]["path"] = path

    return distance_results



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tssearch",
    version="0.1.2",
    author="Fraunhofer Portugal",
    description="Library for time series subsequence search",
    long_description=long_description,
    long_description_content_type="text/markdown",
    download_url="https://github.com/fraunhoferportugal/tssearch/archive/refs/tags/v0.1.1.tar.gz",
    package_data={'tssearch': ['distances/distances.json', 'examples/ecg.pickle']},
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['scipy', 'pandas', 'matplotlib', 'numpy'],
)



import importlib
import inspect
import json
import os
import sys
import warnings
from inspect import getmembers, isfunction

from tssearch.utils.distances_settings import load_json


def add_distance_json(distances_path, json_path):
    """Adds new distance to features.json.
    Parameters
    ----------
    distances_path: string
        Personal Python module directory containing new distances implementation.
    json_path: string
        Personal .json file directory containing existing disatnces from TSSEARCH.
        New customised distances will be added to file in this directory.
    """

    sys.path.append(distances_path[: -len(distances_path.split(os.sep)[-1]) - 1])
    exec("import " + distances_path.split(os.sep)[-1][:-3])

    # Reload module containing the new features
    importlib.reload(sys.modules[distances_path.split(os.sep)[-1][:-3]])
    exec("import " + distances_path.split(os.sep)[-1][:-3] + " as pymodule")

    # Functions from module containing the new features
    functions_list = [o for o in getmembers(locals()["pymodule"]) if isfunction(o[1])]
    function_names = [fname[0] for fname in functions_list]

    # Check if @set_domain was declared on features module
    vset_domain = False

    for fname, f in list(locals()["pymodule"].__dict__.items()):

        if getattr(f, "domain", None) is not None:

            vset_domain = True

            # Access to personal features.json
            feat_json = load_json(json_path)

            # Assign domain and tag
            domain = getattr(f, "domain", None)

            # Feature specifications
            # Description
            if f.__doc__ is not None:
                descrip = f.__doc__.split("\n")[0]
            else:
                descrip = ""
            # Feature usage
            use = "yes"
            # Feature function arguments
            args_name = inspect.getfullargspec(f)[0]

            # Access feature parameters
            if args_name != "":
                # Retrieve default values of arguments
                spec = inspect.getfullargspec(f)
                defaults = dict(zip(spec.args[::-1], (spec.defaults or ())[::-1]))
                defaults.update(spec.kwonlydefaults or {})

                for p in args_name[1:]:
                    if p not in list(defaults.keys()):
                        defaults[p] = None
                if len(defaults) == 0:
                    defaults = ""
            else:
                defaults = ""

            # Settings of new feature
            new_feature = {"description": descrip, "parameters": defaults, "function": fname, "use": use}

            # Check if domain exists
            try:
                feat_json[domain][fname] = new_feature
            except KeyError:
                feat_json[domain] = {fname: new_feature}

            # Write new feature on json file
            with open(json_path, "w") as fout:
                json.dump(feat_json, fout, indent=" ")

            print("Feature " + str(fname) + " was added.")

    if vset_domain is False:
        warnings.warn("No features were added. Please declare @set_domain.", stacklevel=2)



import numpy as np

from tssearch.distances.elastic_distances import dtw
from tssearch.distances.elastic_utils import traceback


def tam(x, y):
    """Calculates the Time Alignment Measurement (TAM) based on an optimal warping path
    between two time series.

    Reference: Folgado et. al, Time Alignment Measurement for Time Series, 2016.

    Parameters
    ----------
    x : nd-array
        Time series x.
    y : nd-array
        Time series y.

    Returns
    -------
    In case ``report=instants`` the number of indexes in advance, delay and phase
    will be returned.
    For ``report=ratios``, the ratio of advance, delay and phase.
    will be returned. In case ``report=distance``, only the TAM will be returned.

    """
    ac = dtw(x, y, report="cost_matrix")

    path = traceback(ac)

    # Delay and advance counting
    delay = len(np.where(np.diff(path[0]) == 0)[0])
    advance = len(np.where(np.diff(path[1]) == 0)[0])

    # Phase counting
    incumbent = np.where((np.diff(path[0]) == 1) * (np.diff(path[1]) == 1))[0]
    phase = len(incumbent)

    # Estimated and reference time series duration.
    len_estimation = path[1][-1]
    len_ref = path[0][-1]

    p_advance = advance * 1.0 / len_ref
    p_delay = delay * 1.0 / len_estimation
    p_phase = phase * 1.0 / np.min([len_ref, len_estimation])

    distance = p_advance + p_delay + (1 - p_phase)
    return distance
